/* ***********************************************
 
	GEAM for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_geam_nn_kernel.cu"
//#include "mublas_geam_tn_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
__host__ int32_t
MUBLAS_GEAM (
	mublasHandle_t* mh,	
	const char transA,
	const char transB,
	const size_t m,
	const size_t n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *beta,
	const TYPE0 *b,
	const size_t ldb,
	TYPE0 *c,
	const size_t ldc
) {
	uint32_t kernel_number = 999;
    void (*kernel) (const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, TYPE0*, const SIZE_T);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->m = m;
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)a%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)b%(128/sizeof(TYPE0)))%16 != 0) ||
		(((long)c%(128/sizeof(TYPE0)))%16 != 0) ||
		(lda%mh->N_T != 0) ||
		(ldb%mh->N_T != 0) ||
		(ldc%mh->N_T != 0)) mh->N_T = 1;
	if (transB == 'N' || transB == 'n') {
		if (transA == 'N' || transA == 'n') {
			kernel_number = 100 + mh->N_T * 10;
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEAM_NN;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEAM_NN;
			mh->func_cache = cudaFuncCachePreferL1;
			#include "mublas_geam_nn.incl"
		}
	}
	switch (kernel_number) {
		case 140: kernel = &MUBLAS_GEAM_NN_DEVICE_v4; break;
		case 120: kernel = &MUBLAS_GEAM_NN_DEVICE_v2; break;
		case 110: kernel = &MUBLAS_GEAM_NN_DEVICE_s1; break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_GEAM)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		cublasOperation_t tranBcublas = CUBLAS_OP_N;
		if (transA == 'T' || transA == 't') tranAcublas = CUBLAS_OP_T;
		if (transA == 'C' || transA == 'c') tranAcublas = CUBLAS_OP_C;
		if (transB == 'T' || transB == 't') tranBcublas = CUBLAS_OP_T;
		if (transB == 'C' || transB == 'c') tranBcublas = CUBLAS_OP_C;
		CUBLAS_GEAM (ch, tranAcublas, tranBcublas, m, n, alpha, a, lda, beta, b, ldb, c, ldc);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------
	
	// ---------------------------------------------
	// 4. Thread-Block Size Adjust and Kernel Launch
	cudaFuncGetAttributes (&mh->funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh->func_cache);
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, *beta, b, ldb, c, ldc);
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

