/* ***********************************************
 
	GEMM-NN for CUDA 

   *********************************************** */   

#define	USE_NC_A_GEMM_NN	1
#define	USE_NC_B_GEMM_NN	1

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMM_NN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_B_GEMM_NN
#define RESTRICT_B __restrict__
#define LOAD1_B LOAD1_NC
#define LOAD2_B LOAD2_NC
#define LOAD4_B LOAD4_NC
#else
#define RESTRICT_B
#define LOAD1_B LOAD1___
#define LOAD2_B LOAD2___
#define LOAD4_B LOAD4___
#endif
#define LOAD1_C LOAD1___
#define LOAD2_C LOAD2___
#define STORE1_C STORE1___
#define STORE2_C STORE2___

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb1x1 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb1x1 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb2x2 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb2x2 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb8x8 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * 8;
	const SIZE_T Bn = mh->Ty * 8;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb8x8 ( mublasHandle_t* mh) {
	const SIZE_T Bm = mh->Tx * 8;
	const SIZE_T Bn = mh->Ty * 8;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_GEMM_NN_rb8x8 ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)8 / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

//__global__ void __launch_bounds__(64, 4) MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell (
__global__ void MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC4
	int32_t i, j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*64+iTx*4;
	const int32_t addry = iBy*64+iTy*8;
	__shared__ TYPE4 As4[8][16+1];
	__shared__ TYPE4 Bs4[8][16+1];
	register TYPE4 Tr4[8][2];
	register TYPE4 Ar4[2];
	register TYPE4 Br4[2];
	register TYPE4 Cr4[2];

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	const TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	#pragma unroll
	for (j = 0; j < 8; j++) 
		#pragma unroll
		for (i = 0; i < 2; i++)
			Tr4[j][i] = ZERO4;

	for (jk = 0; jk < k; jk += 8) {
		LOAD4_A (As4[iTy][iTx+0], addrAg+0);
		LOAD4_A (As4[iTy][iTx+8], addrAg+32);
		LOAD1_B (Br4[0].x, addrBg+0*ldb);
		LOAD1_B (Br4[0].y, addrBg+1*ldb);
		LOAD1_B (Br4[0].z, addrBg+2*ldb);
		LOAD1_B (Br4[0].w, addrBg+3*ldb);
		LOAD1_B (Br4[1].x, addrBg+4*ldb);
		LOAD1_B (Br4[1].y, addrBg+5*ldb);
		LOAD1_B (Br4[1].z, addrBg+6*ldb);
		LOAD1_B (Br4[1].w, addrBg+7*ldb);
		Bs4[iTx][iTy*2+0] = Br4[0];
		Bs4[iTx][iTy*2+1] = Br4[1];
		__syncthreads ();

		#pragma unroll
		for (ki = 0; ki < 8; ki++) {
			Ar4[0] = As4[ki][iTx+0];
			Ar4[1] = As4[ki][iTx+8];
			Br4[0] = Bs4[ki][iTy*2+0];
			Br4[1] = Bs4[ki][iTy*2+1];
			Tr4[0][0] = MAD414 (Ar4[0], Br4[0].x, Tr4[0][0]); 
			Tr4[1][0] = MAD414 (Ar4[0], Br4[0].y, Tr4[1][0]); 
			Tr4[2][0] = MAD414 (Ar4[0], Br4[0].z, Tr4[2][0]); 
			Tr4[3][0] = MAD414 (Ar4[0], Br4[0].w, Tr4[3][0]); 
			Tr4[4][0] = MAD414 (Ar4[0], Br4[1].x, Tr4[4][0]); 
			Tr4[5][0] = MAD414 (Ar4[0], Br4[1].y, Tr4[5][0]); 
			Tr4[6][0] = MAD414 (Ar4[0], Br4[1].z, Tr4[6][0]); 
			Tr4[7][0] = MAD414 (Ar4[0], Br4[1].w, Tr4[7][0]); 
			Tr4[0][1] = MAD414 (Ar4[1], Br4[0].x, Tr4[0][1]); 
			Tr4[1][1] = MAD414 (Ar4[1], Br4[0].y, Tr4[1][1]); 
			Tr4[2][1] = MAD414 (Ar4[1], Br4[0].z, Tr4[2][1]); 
			Tr4[3][1] = MAD414 (Ar4[1], Br4[0].w, Tr4[3][1]); 
			Tr4[4][1] = MAD414 (Ar4[1], Br4[1].x, Tr4[4][1]); 
			Tr4[5][1] = MAD414 (Ar4[1], Br4[1].y, Tr4[5][1]); 
			Tr4[6][1] = MAD414 (Ar4[1], Br4[1].z, Tr4[6][1]); 
			Tr4[7][1] = MAD414 (Ar4[1], Br4[1].w, Tr4[7][1]); 
		}
		__syncthreads ();
		addrAg += 8*lda;
		addrBg += 8;
	}

	#pragma unroll 
	for (j = 0; j < 8; j++) {
		Cr4[0] = (TYPE4&)*(addrCg+ 0);
		Cr4[1] = (TYPE4&)*(addrCg+32);
		Cr4[0] = MAD144 (alpha, Tr4[j][0], MUL14 (beta, Cr4[0]));
		Cr4[1] = MAD144 (alpha, Tr4[j][1], MUL14 (beta, Cr4[1]));
		(TYPE4&)*(addrCg+ 0) = Cr4[0];
		(TYPE4&)*(addrCg+32) = Cr4[1];
		addrCg += ldc;
	}

#endif
}

__global__ void __launch_bounds__(64, 8) MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v2_maxwell (
//__global__ void __launch_bounds__(64, 4) MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v2_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t i, j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*64+iTx*2;
	const int32_t addry = iBy*64+iTy*8;
	__shared__ TYPE2 As2[8][32+1];
	__shared__ TYPE2 Bs2[8][32+1];
	register TYPE2 Tr2[8][4];
	register TYPE2 Ar2[4];
	register TYPE2 Br2[4];
	register TYPE2 Cr2[4];

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	const TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	#pragma unroll
	for (j = 0; j < 8; j++) 
		#pragma unroll
		for (i = 0; i < 4; i++) 
			Tr2[j][i] = ZERO2;

	for (jk = 0; jk < k; jk += 8) {
		LOAD2_A (As2[iTy][iTx+ 0], addrAg+0);
		LOAD2_A (As2[iTy][iTx+ 8], addrAg+16);
		LOAD2_A (As2[iTy][iTx+16], addrAg+32);
		LOAD2_A (As2[iTy][iTx+24], addrAg+48);
		LOAD1_B ((TYPE1&)Br2[0].x, addrBg+0*ldb);
		LOAD1_B ((TYPE1&)Br2[0].y, addrBg+1*ldb);
		LOAD1_B ((TYPE1&)Br2[1].x, addrBg+2*ldb);
		LOAD1_B ((TYPE1&)Br2[1].y, addrBg+3*ldb);
		LOAD1_B ((TYPE1&)Br2[2].x, addrBg+4*ldb);
		LOAD1_B ((TYPE1&)Br2[2].y, addrBg+5*ldb);
		LOAD1_B ((TYPE1&)Br2[3].x, addrBg+6*ldb);
		LOAD1_B ((TYPE1&)Br2[3].y, addrBg+7*ldb);
		Bs2[iTx][iTy*4+0] = Br2[0];
		Bs2[iTx][iTy*4+1] = Br2[1];
		Bs2[iTx][iTy*4+2] = Br2[2];
		Bs2[iTx][iTy*4+3] = Br2[3];
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 8; ki++) {
			Ar2[0] = As2[ki][iTx+ 0];
			Ar2[1] = As2[ki][iTx+ 8];
			Ar2[2] = As2[ki][iTx+16];
			Ar2[3] = As2[ki][iTx+24];
			Br2[0] = Bs2[ki][iTy*4+0];
			Br2[1] = Bs2[ki][iTy*4+1];
			Br2[2] = Bs2[ki][iTy*4+2];
			Br2[3] = Bs2[ki][iTy*4+3];
			#pragma unroll
			for (i = 0; i < 4; i++) {
				Tr2[0][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].x, Tr2[0][i]); 
				Tr2[1][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].y, Tr2[1][i]); 
				Tr2[2][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].x, Tr2[2][i]); 
				Tr2[3][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].y, Tr2[3][i]); 
				Tr2[4][i] = MAD212 (Ar2[i], (TYPE1&)Br2[2].x, Tr2[4][i]); 
				Tr2[5][i] = MAD212 (Ar2[i], (TYPE1&)Br2[2].y, Tr2[5][i]); 
				Tr2[6][i] = MAD212 (Ar2[i], (TYPE1&)Br2[3].x, Tr2[6][i]); 
				Tr2[7][i] = MAD212 (Ar2[i], (TYPE1&)Br2[3].y, Tr2[7][i]); 
			}
		}
		__syncthreads ();
		addrAg += 8*lda;
		addrBg += 8;
	}

	#pragma unroll 
	for (j = 0; j < 8; j++) {
		Cr2[0] = (TYPE2&)*(addrCg+ 0);
		Cr2[1] = (TYPE2&)*(addrCg+16);
		Cr2[2] = (TYPE2&)*(addrCg+32);
		Cr2[3] = (TYPE2&)*(addrCg+48);
		Cr2[0] = MAD122 (LOAD0(alpha), Tr2[j][0], MUL12 (LOAD0(beta), Cr2[0]));
		Cr2[1] = MAD122 (LOAD0(alpha), Tr2[j][1], MUL12 (LOAD0(beta), Cr2[1]));
		Cr2[2] = MAD122 (LOAD0(alpha), Tr2[j][2], MUL12 (LOAD0(beta), Cr2[2]));
		Cr2[3] = MAD122 (LOAD0(alpha), Tr2[j][3], MUL12 (LOAD0(beta), Cr2[3]));
		(TYPE2&)*(addrCg+ 0) = Cr2[0];
		(TYPE2&)*(addrCg+16) = Cr2[1];
		(TYPE2&)*(addrCg+32) = Cr2[2];
		(TYPE2&)*(addrCg+48) = Cr2[3];
		addrCg += ldc;
	}
#endif
}

__global__ void __launch_bounds__(256, 1) MUBLAS_GEMM_NN_DEVICE_tb16x16_rb8x8_v2_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb8x8_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t i, j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*128+iTx*2;
	const int32_t addry = iBy*128+iTy*8;
	__shared__ TYPE2 As2[16][64+1];
	__shared__ TYPE2 Bs2[16][64+1];
	register TYPE2 Tr2[8][4];
	register TYPE2 Ar2[4];
	register TYPE2 Br2[4];
	register TYPE2 Cr2[4];

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	const TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	#pragma unroll
	for (j = 0; j < 8; j++) 
		#pragma unroll
		for (i = 0; i < 4; i++) 
			Tr2[j][i] = ZERO2;

	for (jk = 0; jk < k; jk += 16) {
		LOAD2_A (As2[iTy][iTx+ 0], addrAg+0);
		LOAD2_A (As2[iTy][iTx+16], addrAg+32);
		LOAD2_A (As2[iTy][iTx+32], addrAg+64);
		LOAD2_A (As2[iTy][iTx+48], addrAg+96);

		// matrix B transpose on register
		LOAD1_B ((TYPE1&)Br2[0].x, addrBg+0*ldb);
		LOAD1_B ((TYPE1&)Br2[0].y, addrBg+1*ldb);
		LOAD1_B ((TYPE1&)Br2[1].x, addrBg+2*ldb);
		LOAD1_B ((TYPE1&)Br2[1].y, addrBg+3*ldb);
		LOAD1_B ((TYPE1&)Br2[2].x, addrBg+4*ldb);
		LOAD1_B ((TYPE1&)Br2[2].y, addrBg+5*ldb);
		LOAD1_B ((TYPE1&)Br2[3].x, addrBg+6*ldb);
		LOAD1_B ((TYPE1&)Br2[3].y, addrBg+7*ldb);

		Bs2[iTx][iTy*4+0] = Br2[0];
		Bs2[iTx][iTy*4+1] = Br2[1];
		Bs2[iTx][iTy*4+2] = Br2[2];
		Bs2[iTx][iTy*4+3] = Br2[3];
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar2[0] = As2[ki][iTx+ 0];
			Ar2[1] = As2[ki][iTx+16];
			Ar2[2] = As2[ki][iTx+32];
			Ar2[3] = As2[ki][iTx+48];
			Br2[0] = Bs2[ki][iTy*4+0];
			Br2[1] = Bs2[ki][iTy*4+1];
			Br2[2] = Bs2[ki][iTy*4+2];
			Br2[3] = Bs2[ki][iTy*4+3];
			#pragma unroll 
			for (i = 0; i < 4; i++) {
				Tr2[0][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].x, Tr2[0][i]); 
				Tr2[1][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].y, Tr2[1][i]); 
				Tr2[2][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].x, Tr2[2][i]); 
				Tr2[3][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].y, Tr2[3][i]); 
				Tr2[4][i] = MAD212 (Ar2[i], (TYPE1&)Br2[2].x, Tr2[4][i]); 
				Tr2[5][i] = MAD212 (Ar2[i], (TYPE1&)Br2[2].y, Tr2[5][i]); 
				Tr2[6][i] = MAD212 (Ar2[i], (TYPE1&)Br2[3].x, Tr2[6][i]); 
				Tr2[7][i] = MAD212 (Ar2[i], (TYPE1&)Br2[3].y, Tr2[7][i]); 
			}
		}
		__syncthreads ();

		addrAg += 16*lda;
		addrBg += 16;
	}

	#pragma unroll 
	for (j = 0; j < 8; j++) {
		Cr2[0] = (TYPE2&)*(addrCg+ 0);
		Cr2[1] = (TYPE2&)*(addrCg+32);
		Cr2[2] = (TYPE2&)*(addrCg+64);
		Cr2[3] = (TYPE2&)*(addrCg+96);
		Cr2[0] = MAD122 (LOAD0(alpha), Tr2[j][0], MUL12 (LOAD0(beta), Cr2[0]));
		Cr2[1] = MAD122 (LOAD0(alpha), Tr2[j][1], MUL12 (LOAD0(beta), Cr2[1]));
		Cr2[2] = MAD122 (LOAD0(alpha), Tr2[j][2], MUL12 (LOAD0(beta), Cr2[2]));
		Cr2[3] = MAD122 (LOAD0(alpha), Tr2[j][3], MUL12 (LOAD0(beta), Cr2[3]));
		(TYPE2&)*(addrCg+ 0) = Cr2[0];
		(TYPE2&)*(addrCg+32) = Cr2[1];
		(TYPE2&)*(addrCg+64) = Cr2[2];
		(TYPE2&)*(addrCg+96) = Cr2[3];
		addrCg += ldc;
	}

#endif
}

#ifdef PREC_DB
__device__ __forceinline__ void ld_global_nc_v2_f64_db (cudfreal2 &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#endif
#else
	reg = (cudfreal2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_f64_db (cudfreal &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}
#endif

__global__ void __launch_bounds__(256, 6) MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
#ifdef PREC_DB
	const cudfreal * RESTRICT_A Abg,
	const cudfreal * RESTRICT_B Bbg,
#endif
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*32+iTx*2;
	const int32_t addry = iBy*32+iTy*2;
	__shared__ TYPE2 As2[16][16+1];
	__shared__ TYPE2 Bs2[16][16+1];
	register TYPE2 Tr2[2];
	register TYPE2 Ar2;
	register TYPE2 Br2;
	register TYPE2 Cr2;

	#ifdef PREC_DB
	const TYPE1 * addrAg = &Abg[iTy*lda + addrx];
	const TYPE1 * addrBg = &Bbg[addry*ldb + iTx];
	#else
	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	#endif
	TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	Tr2[0] = ZERO2;
	Tr2[1] = ZERO2;

	for (jk = 0; jk < k; jk += 16) {
		#ifdef PREC_DB
		ld_global_nc_v2_f64_db (As2[iTy][iTx], addrAg);
		ld_global_nc_f64_db ((TYPE1&)Br2.x, addrBg+0*ldb);
		ld_global_nc_f64_db ((TYPE1&)Br2.y, addrBg+1*ldb);
		#else
		LOAD2_A (As2[iTy][iTx], addrAg);
		LOAD1_B ((TYPE1&)Br2.x, addrBg+0*ldb);
		LOAD1_B ((TYPE1&)Br2.y, addrBg+1*ldb);
		#endif
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();

		addrAg += 16*lda;
		addrBg += 16;
	}

	#pragma unroll 
	for (j = 0; j < 2; j++) {
		LOAD2_C (Cr2, addrCg);
		Cr2 = MAD122 (LOAD0(alpha), Tr2[j], MUL12 (LOAD0(beta), Cr2));
		STORE2_C (addrCg, Cr2);
		addrCg += ldc;
	}

#endif
}

/*
__global__ void __launch_bounds__(64, 4) MUBLAS_GEMM_NN_DEVICE_tb8x8_rb2x2_v2_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*16+iTx*2;
	const int32_t addry = iBy*16+iTy*2;
	__shared__ TYPE2 As2[8][8+1];
	__shared__ TYPE2 Bs2[8][8+1];
	register TYPE2 Tr2[2];
	register TYPE2 Ar2;
	register TYPE2 Br2;
	register TYPE2 Cr2;

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	Tr2[0] = ZERO2;
	Tr2[1] = ZERO2;

	for (jk = 0; jk < k; jk += 8) {
		LOAD2_A (As2[iTy][iTx], addrAg);
		LOAD1_B ((TYPE1&)Br2.x, addrBg+0*ldb);
		LOAD1_B ((TYPE1&)Br2.y, addrBg+1*ldb);
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 8; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();

		addrAg += 8*lda;
		addrBg += 8;
	}

	#pragma unroll 
	for (j = 0; j < 2; j++) {
		LOAD2_C (Cr2, addrCg);
		Cr2 = MAD122 (LOAD0(alpha), Tr2[j], MUL12 (LOAD0(beta), Cr2));
		STORE2_C (addrCg, Cr2);
		addrCg += ldc;
	}

#endif
}
*/
/*
__global__ void __launch_bounds__(256, 2) MUBLAS_GEMM_NN_DEVICE_tb16x16_rb4x4_v2_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb4x4_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t i, j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*64+iTx*2;
	const int32_t addry = iBy*64+iTy*4;
	__shared__ TYPE2 As2[16][32+1];
	__shared__ TYPE2 Bs2[16][32+1];
	register TYPE2 Tr2[4][2];
	register TYPE2 Ar2[2];
	register TYPE2 Br2[2];
	register TYPE2 Cr2[2];

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	#pragma unroll
	for (j = 0; j < 4; j++) 
		#pragma unroll
		for (i = 0; i < 2; i++) 
			Tr2[j][i] = ZERO2;

	for (jk = 0; jk < k; jk += 16) {
		LOAD2_A (As2[iTy][iTx+ 0], addrAg+0);
		LOAD2_A (As2[iTy][iTx+16], addrAg+32);

		// matrix B transpose on register
		LOAD1_B ((TYPE1&)Br2[0].x, addrBg+0*ldb);
		LOAD1_B ((TYPE1&)Br2[0].y, addrBg+1*ldb);
		LOAD1_B ((TYPE1&)Br2[1].x, addrBg+2*ldb);
		LOAD1_B ((TYPE1&)Br2[1].y, addrBg+3*ldb);
		Bs2[iTx][iTy*2+0] = Br2[0];
		Bs2[iTx][iTy*2+1] = Br2[1];
		__syncthreads ();
	
		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar2[0] = As2[ki][iTx+ 0];
			Ar2[1] = As2[ki][iTx+16];
			Br2[0] = Bs2[ki][iTy*2+0];
			Br2[1] = Bs2[ki][iTy*2+1];

			#pragma unroll 
			for (i = 0; i < 2; i++) {
				Tr2[0][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].x, Tr2[0][i]); 
				Tr2[1][i] = MAD212 (Ar2[i], (TYPE1&)Br2[0].y, Tr2[1][i]); 
				Tr2[2][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].x, Tr2[2][i]); 
				Tr2[3][i] = MAD212 (Ar2[i], (TYPE1&)Br2[1].y, Tr2[3][i]); 
			}
		}
		__syncthreads ();

		addrAg += 16*lda;
		addrBg += 16;
	}

	#pragma unroll 
	for (j = 0; j < 4; j++) {
		LOAD2_C (Cr2[0], addrCg);
		LOAD2_C (Cr2[1], addrCg+32);
		Cr2[0] = MAD122 (LOAD0(alpha), Tr2[j][0], MUL12 (LOAD0(beta), Cr2[0]));
		Cr2[1] = MAD122 (LOAD0(alpha), Tr2[j][1], MUL12 (LOAD0(beta), Cr2[1]));
		STORE2_C (addrCg+ 0, Cr2[0]);
		STORE2_C (addrCg+32, Cr2[1]);
		addrCg += ldc;
	}


#endif
}
*/

/* ***********************************************
	(simple) kernel=s1, trans=N
   *********************************************** */   

__global__ void MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v1_simple (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC1
	SIZE_T jk, ik, iTh, ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	__shared__ TYPE1 Ts[1024];
	const SIZE_T ofs = 512;
	TYPE1 Tr[8][8], Cr;

	#pragma unroll
	for (iy = 0; iy < 8; iy++) 
		#pragma unroll
		for (ix = 0; ix < 8; ix++) 
			Tr[iy][ix] = ZERO1;

	for (jk = 0; jk < k; jk += 8) {
		#pragma unroll
		for (ix = 0; ix < 8; ix++) 
			LOAD1_A ( Ts[iTy*64+(8*ix+iTx)], &Ag[(jk+iTy)*lda + 64*iBx+8*ix+iTx]);
		#pragma unroll
		for (iy = 0; iy < 8; iy++) 
			LOAD1_B (Ts[ofs+(8*iTy+iy)*8+iTx], &Bg[(64*iBy+8*iTy+iy)*ldb + jk+iTx]);
		__syncthreads ();

		#pragma unroll
		for (iy = 0; iy < 8; iy++) 
			#pragma unroll
			for (ix = 0; ix < 8; ix++) 
				#pragma unroll
				for (ik = 0; ik < 8; ik++) 
					Tr[iy][ix] = MAD1 (Ts[8*ix+iTx+ik*64], Ts[ofs+(8*iTy+iy)*8+ik], Tr[iy][ix]); 
		__syncthreads ();
	}

	#pragma unroll
	for (iy = 0; iy < 8; iy++) {
		#pragma unroll
		for (ix = 0; ix < 8; ix++) {
			iTh = (64*iBy+8*iTy+iy) * ldc + (64*iBx+8*ix+iTx);
			LOAD1_C (Cr, &Cg[iTh]);
			Cr = MAD1 (LOAD0(alpha), Tr[iy][ix], MUL1 (LOAD0(beta), Cr));
			STORE1_C (&Cg[iTh], Cr);
		}
	}
#endif
}

/* ***********************************************
	(full) kernel=s1, trans=N
   *********************************************** */   

__global__ void __launch_bounds__(256, 6) MUBLAS_GEMM_NN_DEVICE_tb16x16_rb1x1_s1_maxwell (
//__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
#ifdef PREC_DB
	const cudfreal * RESTRICT_A Abg,
	const cudfreal * RESTRICT_B Bbg,
#endif
	TYPE0 *Cg,
	const SIZE_T ldc
) {
	int32_t jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*16+iTx;
	const int32_t addry = iBy*16+iTy;
	__shared__ TYPE1 As1[16][16+1];
	__shared__ TYPE1 Bs1[16][16+1];
	register TYPE1 Ar1, Br1, Cr1, Tr1;

	TYPE0 * addrCg = &Cg[addry*ldc + addrx];
	Tr1 = ZERO1;

	for (jk = 0; jk < k-16; jk += 16) {
		#ifdef PREC_DB
		ld_global_nc_f64_db (As1[iTy][iTx], &Abg[min(k-1,jk+iTy)*lda + min(m-1,addrx)]);
		ld_global_nc_f64_db (Br1, &Bbg[min(n-1,addry)*ldb + min(k-1,iTx+jk)]);
		#else
		LOAD1_A (As1[iTy][iTx], &Ag[min(k-1,jk+iTy)*lda + min(m-1,addrx)]);
		LOAD1_B (Br1, &Bg[min(n-1,addry)*ldb + min(k-1,iTx+jk)]);
		#endif
		Bs1[iTx][iTy] = Br1;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	}

	{
		if (jk+iTy < k && addrx < m) {
			#ifdef PREC_DB
			ld_global_nc_f64_db (As1[iTy][iTx], &Abg[(jk+iTy)*lda + (addrx)]);
			#else
			LOAD1_A (As1[iTy][iTx], &Ag[(jk+iTy)*lda + (addrx)]);
			#endif
		} else {
			As1[iTy][iTx] = ZERO1;
		}
		if (jk+iTx < k && addry < n) {
			#ifdef PREC_DB
			ld_global_nc_f64_db (Br1, &Bbg[(addry)*ldb + (iTx+jk)]);
			#else
			LOAD1_B (Br1, &Bg[(addry)*ldb + (iTx+jk)]);
			#endif
			Bs1[iTx][iTy] = Br1;
		} else {
			Bs1[iTx][iTy] = ZERO1;
		}
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	}

//	if (addrx < m && addry < n) {
		LOAD1_C (Cr1, addrCg);
		Cr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Cr1));
		STORE1_C (addrCg, Cr1);
//	}
}



/* ***********************************************
	kernel=s1, trans=N
   *********************************************** */   
   /*
__global__ void MUBLAS_GEMM_NN_DEVICE_s1 (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC1
	SIZE_T i, j, kb, ki, tid;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T addrx = iBx*VNBK + iTx*VEC;
	const SIZE_T addry = iBy*VNBK + iTy*VNBK/VNTH;
	const SIZE_T pdoffset = m%VEC;
	__shared__ VTYPEIN As[VNTH][VNBK/VEC];
	__shared__ TYPEIN1 Bs[VNTH][VNBK+1];
	VTYPEIN Tr[VNBK/VNTH][VNBK/VNTH/VEC];
	VTYPEIN Ar[VNBK/VNTH/VEC];
	TYPEIN1 Br[VNBK/VNTH];

	#pragma unroll
	for (j = 0; j < VNBK/VNTH; j++) {
		#pragma unroll
		for (i = 0; i < VNBK/VNTH/VEC; i++) 
			Tr[j][i] = VZEROIN;
	}

	for (kb = 0; kb < k - VNTH; kb += VNTH) {
		#pragma unroll
		for (i = 0; i < VNBK/VNTH/VEC; i++) 
			As[iTy][iTx+i*VNTH] = VLOAD ((VTYPE&) A[min(k-1,kb+iTy) * lda + min(m-pdoffset,addrx+i*VNTH*VEC)]);
		#pragma unroll
		for (j = 0; j < VNBK/VNTH; j++)
			Bs[iTx][iTy*VNBK/VNTH+j] = LOAD1 (B[min(n-1,addry+j) * ldb + min(k-1,kb+iTx)]);
		__syncthreads ();

		#pragma unroll
		for (ki = 0; ki < VNTH; ki++) {
			#pragma unroll
			for (i = 0; i < VNBK/VNTH/VEC; i++) 
				Ar[i] = As[ki][iTx+i*VNTH];
			#pragma unroll
			for (i = 0; i < VNBK/VNTH; i++) 
				(VTYPE&) Br[i*VEC] = (VTYPE&) Bs[ki][iTy*VNBK/VNTH+i*VEC];
			#pragma unroll
			for (j = 0; j < VNBK/VNTH; j++) {
				#pragma unroll
				for (i = 0; i < VNBK/VNTH/VEC; i++) 
					Tr[j][i] = MAD1 (Ar[i], Br[j], Tr[j][i]); 
			}
		}
		__syncthreads ();
	}

	// last one block
	#pragma unroll
	for (i = 0; i < VNBK/VNTH/VEC; i++) {
		if (kb+iTy < k && addrx+i*VNTH*VEC < m)
			As[iTy][iTx+i*VNTH] = VLOAD ((VTYPE&) A[(kb+iTy) * lda + addrx+i*VNTH*VEC]);
			//here: do not use the following pointer redirecting code. illegal memory access occurs in some conditions only on DF-operation
			//As[iTy][iTx+i*VNTH] = VLOAD ((VTYPE&) A[(kb+iTy) * lda + min(m-pdoffset,addrx+i*VNTH*VEC)]);
		else
			As[iTy][iTx+i*VNTH] = VZEROIN;
	}
	#pragma unroll
	for (j = 0; j < VNBK/VNTH; j++) {
		if (kb+iTx < k && addry+j < n)
			Bs[iTx][iTy*VNBK/VNTH+j] = LOAD1 (B[(addry+j) * ldb + (kb+iTx)]);
		else
			Bs[iTx][iTy*VNBK/VNTH+j] = ZEROIN1;
	}
	__syncthreads ();

	#pragma unroll 
	for (ki = 0; ki < VNTH; ki++) {
		#pragma unroll
		for (i = 0; i < VNBK/VNTH/VEC; i++) 
			Ar[i] = As[ki][iTx+i*VNTH];
		#pragma unroll
		for (i = 0; i < VNBK/VNTH/VEC; i++) 
			(VTYPE&) Br[i*VEC] = (VTYPE&) Bs[ki][iTy*VNBK/VNTH+i*VEC];
		#pragma unroll
		for (j = 0; j < VNBK/VNTH; j++) {
			#pragma unroll
			for (i = 0; i < VNBK/VNTH/VEC; i++) 
				Tr[j][i] = MAD1 (Ar[i], Br[j], Tr[j][i]); 
		}
	}
	__syncthreads ();

	#pragma unroll
	for (j = 0; j < VNBK/VNTH; j++) {
		#pragma unroll
		for (i = 0; i < VNBK/VNTH/VEC; i++) {
			if (addrx+i*VNTH*VEC < m && addry+j < n) {
				tid = (addry+j) * ldc + (addrx+i*VNTH*VEC);
				(VTYPE&) C[tid] = VSTORE (MAD1 (LOAD1 (alpha), Tr[j][i], MUL1 (LOAD1 (beta), VLOAD ((VTYPE&) C[tid]))));
			}
		}
	}
#endif
}
*/
