#ifndef PREC_DB_H
#define PREC_DB_H

#include "cudf.h"

#ifdef PREC_DB
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasDBgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasDBgemvMgpu
	#define MUBLAS_CSRMV mublasDBcsrmv
	#define MUBLAS_GEAM mublasDBgeam
	#define MUBLAS_MATCONV mublasDtoDBmatconv
	#define MUBLAS_GEMM mublasDBgemm
	#define MUBLAS_GEMV mublasDBgemv
	#define MUBLAS_SYMV mublasDBsymv
	#define MUBLAS_TRMV mublasDBtrmv
	#define MUBLAS_AXPY mublasDBaxpy
	#define MUBLAS_DOT  mublasDBdot
	#define MUBLAS_SCAL mublasDBscal
	#define MUBLAS_COPY mublasDBcopy
//	#define CUBLAS_GEMM	
//	#define CUBLAS_GEMV 
//	#define CUBLAS_SYMV
//	#define CUBLAS_TRMV 
//	#define CUBLAS_AXPY 
//	#define CUBLAS_DOT
//	#define CUBLAS_SCAL 
//	#define CUBLAS_COPY 
	#define TYPE0 double
	#define ZERO0 0.
	#define ONE0 1.
	#define LOAD0(A) double_to_cudfreal(A)

	// vec1
	#define VEC1 1
	#define TYPE1 cudfreal
	#define ZERO1 make_cudfreal(0.,0.)
	#define LOAD1___ ld_global_f64
	#define LOAD1_NC ld_global_nc_f64
	#define STORE1___ st_global_f64

	#define ADD1 CUBADD
	#define MUL1 CUBMUL
	#define MAD1 CUBMAD

	// vec2
	#define VEC2 1
	#define TYPE2 cudfreal2
	#define ZERO2 make_cudfreal2(make_cudfreal(0.,0.),make_cudfreal(0.,0.))
	#define LOAD2___ ld_global_v2_f64
	#define LOAD2_NC ld_global_nc_v2_f64
	#define STORE2___ st_global_v2_f64

	#define ADD2 cuBadd2
	#define MUL12 cuBmul12
	#define MAD122 cuBmad122
	#define MAD212 cuBmad212
	#define MAD2 cuBmad2

	// vec4
	#define VEC4 0

	// DF-algorithm 
	#define CUBADD cuBaddV2
	#define CUBSUB cuBsubV2
	#define CUBMUL cuBmulV3
	#define CUBMAD cuBmuladd
#endif

// === Vector =====================================================
struct __align__(16) cudfreal2
{
	cudfreal x;
	cudfreal y;
};

__device__ __forceinline__ cudfreal2
make_cudfreal2
	(const cudfreal x, const cudfreal y)
{
	cudfreal2 v;
	v.x.x = x.x;
	v.x.y = x.y;
	v.y.x = y.x;
	v.y.y = y.y;
	return v;
}

__host__ __device__ __forceinline__ bool operator== (const cudfreal a, const cudfreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cudfreal
cuBmuladd
	(const cudfreal a, const cudfreal b, const cudfreal c)
{
	return CUBADD(CUBMUL(a, b), c);
}

__device__ __forceinline__ cudfreal2
cuBadd2
	(const cudfreal2 a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBADD(make_cudfreal(a.x.x,a.x.y),make_cudfreal(b.x.x,b.x.y)), CUBADD(make_cudfreal(a.y.x,a.y.y),make_cudfreal(b.y.x,b.y.y)));
}

__device__ __forceinline__ cudfreal2
cuBmul12
	(const cudfreal a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBMUL(a,make_cudfreal(b.x.x,b.x.y)), CUBMUL(a,make_cudfreal(b.y.x,b.y.y)));
}

__device__ __forceinline__ cudfreal2
cuBmad2
	(const cudfreal2 a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x.x,a.x.y),make_cudfreal(b.x.x,b.x.y),make_cudfreal(c.x.x,c.x.y)), CUBMAD(make_cudfreal(a.y.x,a.y.y),make_cudfreal(b.y.x,b.y.y),make_cudfreal(c.y.x,c.y.y)));
}

__device__ __forceinline__ cudfreal2
cuBmad122
	(const cudfreal a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(a,make_cudfreal(b.x.x,b.x.y),make_cudfreal(c.x.x,c.x.y)), CUBMAD(a,make_cudfreal(b.y.x,b.y.y),make_cudfreal(c.y.x,c.y.y)));
}

__device__ __forceinline__ cudfreal2
cuBmad212
	(const cudfreal2 a, const cudfreal b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x.x,a.x.y),b,make_cudfreal(c.x.x,c.x.y)), CUBMAD(make_cudfreal(a.y.x,a.y.y),b,make_cudfreal(c.y.x,c.y.y)));
}

__device__ __forceinline__ cudfreal2
double2_to_cudfreal2
	(const double2 d2)
{
	return make_cudfreal2 (double_to_cudfreal(d2.x), double_to_cudfreal(d2.y));
}

__device__ __forceinline__ double2
cudfreal2_to_double2
	(const cudfreal2 df2)
{
	return make_double2 (__dadd_rn ((double)df2.x.x, (double)df2.x.y), __dadd_rn ((double)df2.y.x, (double)df2.y.y));
}
// =================================================================

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cudfreal &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = CUBADD (val, make_cudfreal (__shfl_xor(val.x, n, WrpSize), __shfl_xor(val.y, n, WrpSize)));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_nc_v2_f64 (cudfreal2 &reg, const double *addr) {
#if USE_INLINE_ASM
	register double2 rt;
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f64 {%0,%1}, [%2];": "=d"(rt.x), "=d"(rt.y): "l"(addr));
	#else
	asm ("ld.global.v2.f64 {%0,%1}, [%2];": "=d"(rt.x), "=d"(rt.y): "l"(addr));
	#endif
	reg = make_cudfreal2 (double_to_cudfreal(rt.x), double_to_cudfreal(rt.y));
#else
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	reg = make_cudfreal2 (double_to_cudfreal((double)__ldg(addr)), double_to_cudfreal((double)__ldg(addr+1)));
	#else
	reg = make_cudfreal2 (double_to_cudfreal(*addr), double_to_cudfreal(*(addr+1)));
	#endif
#endif
}

__device__ __forceinline__ void ld_global_v2_f64 (cudfreal2 &reg, const double *addr) {
#if USE_INLINE_ASM
	register double2 rt;
	asm ("ld.global.v2.f64 {%0,%1}, [%2];": "=d"(rt.x), "=d"(rt.y): "l"(addr));
	reg = make_cudfreal2 (double_to_cudfreal(rt.x), double_to_cudfreal(rt.y));
#else
	reg = make_cudfreal2 (double_to_cudfreal((double&)*addr), double_to_cudfreal((double&)*(addr+1)));
#endif
}

__device__ __forceinline__ void ld_global_nc_f64 (cudfreal &reg, const double *addr) {
#if USE_INLINE_ASM
	register double rt;
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.f64 %0, [%1];": "=d"(rt): "l"(addr));
	#else
	asm ("ld.global.f64 %0, [%1];": "=d"(rt): "l"(addr));
	#endif
	reg = double_to_cudfreal(rt);
#else
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	reg = double_to_cudfreal(__ldg(addr));
	#else
	reg = double_to_cudfreal(*addr);
	#endif
#endif
}

__device__ __forceinline__ void ld_global_f64 (cudfreal &reg, const double *addr) {
#if USE_INLINE_ASM
	register double rt;
	asm ("ld.global.f64 %0, [%1];": "=d"(rt): "l"(addr));
	reg = double_to_cudfreal(rt);
#else
	reg = double_to_cudfreal(*addr);
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v2_f64 (double *addr, const cudfreal2 reg) {
#if USE_INLINE_ASM
	register double2 rt;
	rt.x = cudfreal_to_double(make_cudfreal(reg.x,reg.y));
	rt.y = cudfreal_to_double(make_cudfreal(reg.z,reg.w));
	asm ("st.global.v2.f64 [%0], {%1,%2};":: "l"(addr), "d"(rt.x), "d"(rt.y));
#else
	(double2&)*addr = make_double2(cudfreal_to_double(make_cudfreal(reg.x.x,reg.x.y)),cudfreal_to_double(make_cudfreal(reg.y.x,reg.y.y)));
#endif
}

__device__ __forceinline__ void st_global_f64 (double *addr, const cudfreal reg) {
#if USE_INLINE_ASM
	register double rt;
	rt = cudfreal_to_double(reg);
	asm ("st.global.f64 [%0], %1;":: "l"(addr), "d"(rt));
#else
	*addr = cudfreal_to_double(reg);
#endif
}

// =================================================================


// === Texture =====================================================
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_a;
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_b;

static __forceinline__ __device__ cudfreal
tex_a(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_a, i);
	return make_cudfreal (v.x, v.y);
}

static __forceinline__ __device__ cudfreal
tex_b(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_b, i);
	return make_cudfreal (v.x, v.y);
}
// =================================================================

#endif
