#ifndef PREC_S_H
#define PREC_S_H

#ifdef PREC_S
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasSgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasSgemvMgpu
	#define MUBLAS_CSRMV mublasScsrmv
	#define MUBLAS_GEAM mublasSgeam
	#define MUBLAS_GEMM mublasSgemm
	#define MUBLAS_GEMMIA mublasSgemmIA
	#define MUBLAS_GEMV mublasSgemv
	#define MUBLAS_SYMV mublasSsymv
	#define MUBLAS_TRMV mublasStrmv
	#define MUBLAS_AXPY mublasSaxpy
	#define MUBLAS_DOT  mublasSdot
	#define MUBLAS_SCAL mublasSscal
	#define MUBLAS_COPY mublasScopy
	#define CUSPARSE_CSRMV cusparseScsrmv
	#define CUBLAS_GEAM cublasSgeam
	#define CUBLAS_GEMM cublasSgemm
	#define CUBLAS_GEMV cublasSgemv
	#define CUBLAS_SYMV cublasSsymv
	#define CUBLAS_TRMV cublasStrmv
	#define CUBLAS_AXPY cublasSaxpy
	#define CUBLAS_DOT  cublasSdot
	#define CUBLAS_SCAL cublasSscal
	#define CUBLAS_COPY cublasScopy
	#define TYPE0 float
	#define ZERO0 0.
	#define ONE0 1.
	#define LOAD0(A) (A)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1(A,B) ((A)+(B))
	#define MUL1(A,B) ((A)*(B))
	#define MAD1(A,B,C) (((A)*(B))+(C))
	#define LOAD1___ ld_global_f32
	#define LOAD1_NC ld_global_nc_f32
	#define STORE1___ st_global_f32
	#define STORE1_ATOMIC st_global_atomic_f32

	// vec2
	#define VEC2 1
	#define TYPE2 float2
	#define ZERO2 make_float2(0.,0.)
	#define ADD2 cuSadd2
	#define MUL12 cuSmul12
	#define MAD122 cuSmad122
	#define MAD212 cuSmad212
	#define MAD2 cuSmad2
	#define LOAD2___ ld_global_v2_f32
	#define LOAD2_NC ld_global_nc_v2_f32
	#define STORE2___ st_global_v2_f32
	#define STORE2_ATOMIC st_global_atomic_v2_f32

	// vec4
	#define VEC4 1
	#define TYPE4 float4
	#define ZERO4 make_float4(0.,0.,0.,0.)
	#define ADD4 cuSadd4
	#define MUL14 cuSmul14
	#define MUL14RU cuSmul14ru
	#define MUL14RD cuSmul14rd
	#define MAD144 cuSmad144
	#define MAD144RU cuSmad144ru
	#define MAD144RD cuSmad144rd
	#define MAD414 cuSmad414
	#define MAD414RU cuSmad414ru
	#define MAD414RD cuSmad414rd
	#define MAD4 cuSmad4
	#define LOAD4___ ld_global_v4_f32
	#define LOAD4_NC ld_global_nc_v4_f32
	#define STORE4___ st_global_v4_f32
	#define STORE4_ATOMIC st_global_atomic_v4_f32

#endif

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (float &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val += __shfl_xor(val, n, WrpSize);
#endif
}
// =================================================================

// === Load ========================================================
// ld_global_f32
__device__ __forceinline__ void ld_global_v4_f32 (float4 &reg, const float *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
#else
	reg = (float4&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_v2_f32 (float2 &reg, const float *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
#else
	reg = (float2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_f32 (float &reg, const float *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.f32 %0, [%1];": "=f"(reg): "l"(addr));
#else
	reg = *addr;
#endif
}

// ld_global_nc_f32
__device__ __forceinline__ void ld_global_nc_v4_f32 (float4 &reg, const float *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#endif
#else
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	reg = __ldg((float4*)addr);
	#else
	reg = (float4&)*addr;
	#endif
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f32 (float2 &reg, const float *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#endif
#else
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	reg = __ldg((float2*)addr);
	#else
	reg = (float2&)*addr;
	#endif
#endif
}

__device__ __forceinline__ void ld_global_nc_f32 (float &reg, const float *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.f32 %0, [%1];": "=f"(reg): "l"(addr));
	#else
	asm ("ld.global.f32 %0, [%1];": "=f"(reg): "l"(addr));
	#endif
#else
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	reg = __ldg(addr);
	#else
	reg = *addr;
	#endif
#endif
}

// ld_shared_f32
__device__ __forceinline__ void ld_shared_v4_f32 (float4 &reg, const float4 *addr) {
#if USE_INLINE_ASM
	asm ("ld.shared.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
#else
	reg = *addr;
#endif
}

// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v4_f32 (float *addr, const float4 reg) {
#if USE_INLINE_ASM
	asm ("st.global.v4.f32 [%0], {%1,%2,%3,%4};":: "l"(addr), "f"(reg.x), "f"(reg.y), "f"(reg.z), "f"(reg.w));
#else
	(float4&)*addr = reg;
#endif
}

__device__ __forceinline__ void st_global_v2_f32 (float *addr, const float2 reg) {
#if USE_INLINE_ASM
	asm ("st.global.v2.f32 [%0], {%1,%2};":: "l"(addr), "f"(reg.x), "f"(reg.y));
#else
	(float2&)*addr = reg;
#endif
}

__device__ __forceinline__ void st_global_f32 (float *addr, const float reg) {
#if USE_INLINE_ASM
	asm ("st.global.f32 [%0], %1;":: "l"(addr), "f"(reg));
#else
	*addr = reg;
#endif
}

__device__ __forceinline__ void st_global_atomic_v4_f32 (float *addr, const float4 reg) {
	atomicAdd (addr+0, reg.x);
	atomicAdd (addr+4, reg.y);
	atomicAdd (addr+8, reg.z);
	atomicAdd (addr+16, reg.w);
}
__device__ __forceinline__ void st_global_atomic_v2_f32 (float *addr, const float2 reg) {
	atomicAdd (addr+0, reg.x);
	atomicAdd (addr+4, reg.y);
}
__device__ __forceinline__ void st_global_atomic_f32 (float *addr, const float reg) {
	atomicAdd (addr, reg);
}
// =================================================================

// === Vector ======================================================
__device__ __forceinline__ float2
cuSadd2
	(const float2 a, const float2 b)
{
	return make_float2 (a.x+b.x, a.y+b.y);
}

__device__ __forceinline__ float2
cuSmul12
	(const float a, const float2 b)
{
	return make_float2 (a*b.x, a*b.y);
}

__device__ __forceinline__ float2
cuSmad122
	(const float a, const float2 b, const float2 c)
{
	return make_float2 (a*b.x+c.x, a*b.y+c.y);
}

__device__ __forceinline__ float2
cuSmad212
	(const float2 a, const float b, const float2 c)
{
	return make_float2 (a.x*b+c.x, a.y*b+c.y);
}

__device__ __forceinline__ float2
cuSmad2
	(const float2 a, const float2 b, const float2 c)
{
	return make_float2 (a.x*b.x+c.x, a.y*b.y+c.y);
}

__device__ __forceinline__ float4
cuSadd4
	(const float4 a, const float4 b)
{
	return make_float4 (a.x+b.x, a.y+b.y, a.z+b.z, a.w+b.w);
}

__device__ __forceinline__ float4
cuSmul14
	(const float a, const float4 b)
{
	return make_float4 (a*b.x, a*b.y, a*b.z, a*b.w);
}

__device__ __forceinline__ float4
cuSmad144
	(const float a, const float4 b, const float4 c)
{
	return make_float4 (a*b.x+c.x, a*b.y+c.y, a*b.z+c.z, a*b.w+c.w);
}

__device__ __forceinline__ float4
cuSmad414
	(const float4 a, const float b, const float4 c)
{
	return make_float4 (a.x*b+c.x, a.y*b+c.y, a.z*b+c.z, a.w*b+c.w);
}

__device__ __forceinline__ float4
cuSmad4
	(const float4 a, const float4 b, const float4 c)
{
	return make_float4 (a.x*b.x+c.x, a.y*b.y+c.y, a.z*b.z+c.z, a.w*b.w+c.w);
}
// =================================================================

// =================================================================
// Interval-Arithmetic 
__device__ __forceinline__ float4
cuSmad414ru
	(const float4 a, const float b, const float4 c)
{
	return make_float4 (__fmaf_ru (a.x,b,c.x), __fmaf_ru(a.y,b,c.y), __fmaf_ru(a.z,b,c.z), __fmaf_ru(a.w,b,c.w));
}
__device__ __forceinline__ float4
cuSmad414rd
	(const float4 a, const float b, const float4 c)
{
	return make_float4 (__fmaf_rd (a.x,b,c.x), __fmaf_rd(a.y,b,c.y), __fmaf_rd(a.z,b,c.z), __fmaf_rd(a.w,b,c.w));
}

__device__ __forceinline__ float4
cuSmad144ru
	(const float a, const float4 b, const float4 c)
{
	return make_float4 (__fmaf_ru(a,b.x,c.x), __fmaf_ru(a,b.y,c.y), __fmaf_ru(a,b.z,c.z), __fmaf_ru(a,b.w,c.w));
}
__device__ __forceinline__ float4
cuSmad144rd
	(const float a, const float4 b, const float4 c)
{
	return make_float4 (__fmaf_rd(a,b.x,c.x), __fmaf_rd(a,b.y,c.y), __fmaf_rd(a,b.z,c.z), __fmaf_rd(a,b.w,c.w));
}

__device__ __forceinline__ float4
cuSmul14ru
	(const float a, const float4 b)
{
	return make_float4 (__fmul_ru(a,b.x), __fmul_ru(a,b.y), __fmul_ru(a,b.z), __fmul_ru(a,b.w));
}
__device__ __forceinline__ float4
cuSmul14rd
	(const float a, const float4 b)
{
	return make_float4 (__fmul_rd(a,b.x), __fmul_rd(a,b.y), __fmul_rd(a,b.z), __fmul_rd(a,b.w));
}
// =================================================================

#endif
