#include "testing_common.h"
#include "testing_level1.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	size_t word_size_r = sizeof (REFERENCE_TYPE);
	double t0, t1;
	cudaError_t err;

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	print_info1 (&th);
// --------------------------------------------

// library setup ------------------------------
	#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t ha;
	mublasCreate (&ha);
	set_mublas_handler (&th, &ha);
	#elif defined (CUBLAS)
	cublasHandle_t ha;
	cublasCreate (&ha);
	cublasSetPointerMode(ha, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
	GPU_TYPE alpha;
	GPU_TYPE *hst_X, *hst_Y, *hst_Y_t;
	GPU_TYPE *dev_X, *dev_Y;
	REFERENCE_TYPE alpha_r = REFERENCE_ZERO;
	REFERENCE_TYPE *hst_X_r = 0, *hst_Y_r = 0;

	// dim setup
	dim_hst_setup (&th);
	// malloc host memory
	cudaMallocHost ((void **) &hst_X, th.word_size * th.dim_n_hst * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * th.dim_n_hst * th.incy);
	// initialize
	mublasInit2D (1, 1, 0, &alpha);
	mublasInit2D (th.dim_n_hst*th.incx, 1, 0, hst_X);
	mublasInit2D (th.dim_n_hst*th.incy, 1, 0, hst_Y);
	if (th.mode == 'c') {
		// initialize
		cudaMallocHost ((void **) &hst_X_r, word_size_r * th.dim_n_hst * th.incx);
		cudaMallocHost ((void **) &hst_Y_r, word_size_r * th.dim_n_hst * th.incy);
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * th.dim_n_hst * th.incy);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_dev_setup (&th);
	while (1) {
		if (th.dim_n_const == 0 && th.dim_n_dev > th.dim_stop) break;

		cudaMalloc ((void **) &dev_X, th.word_size * th.dim_n_dev * th.incx);
		cudaMalloc ((void **) &dev_Y, th.word_size * th.dim_n_dev * th.incy);

		cublasSetVector(th.dim_n_dev * th.incx, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector(th.dim_n_dev * th.incy, th.word_size, hst_Y, 1, dev_Y, 1);

		if (th.mode != 't') {
			printf ("--\t%zd\t--", (size_t)th.dim_n_dev);
		}
		get_routine_theoretial_performance (&th);
	
		// execution ---------------------------------
		if (th.mode == 'p') {
			ROUTINE_TARGET(ha,th,th.dim_n_dev,alpha,dev_X,dev_Y); // dummy
			#if PERF_AVERAGE
			float sec;
			iter = get_num_iter (&th);
			t0 = gettime ();
			for (i = 0; i < iter; i++) ROUTINE_TARGET(ha,th,th.dim_n_dev,alpha,dev_X,dev_Y);
			t1 = gettime ();
			sec = (t1 - t0) / (float)iter;
			if (th.mode == 'e') printf ("\t%f\t%f", t0, t1);
			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
			#else
			float sec_all_exec[MINLOOP];
			float sec_best;//, sec_worst;
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				t0 = gettime ();
				ROUTINE_TARGET(ha,th,th.dim_n_dev,alpha,dev_X,dev_Y);
				t1 = gettime ();
				sec_all_exec[i] = (t1 - t0);
			}
			sec_best  = get_best  (sec_all_exec);
			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec_best, (th.routine_flops / sec_best) * 1.0e-9, (th.routine_bytes / sec_best) * 1.0e-9, (size_t)iter);
//			sec_worst = get_worst (sec_all_exec);
//			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec_worst, (th.routine_flops / sec_worst) * 1.0e-9, (th.routine_bytes / sec_worst) * 1.0e-9, (size_t)iter);
			#endif
		}
		if (th.mode == 'c' || th.mode == 't') 
			ROUTINE_TARGET(ha,th,th.dim_n_dev,alpha,dev_X,dev_Y);
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') 
			cublasGetVector (th.dim_n_dev * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);

		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			mublasConvert2D (1, 1, 0, &alpha, &alpha_r);
			mublasConvert2D (th.dim_n_hst*th.incx, 1, 0, hst_X, hst_X_r);
			mublasConvert2D (th.dim_n_hst*th.incy, 1, 0, hst_Y, hst_Y_r);
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_X_r, *dev_Y_r;
			cudaMalloc ((void **) &dev_X_r, word_size_r * th.dim_n_dev * th.incx);
			cudaMalloc ((void **) &dev_Y_r, word_size_r * th.dim_n_dev * th.incy);
			cublasSetVector (th.dim_n_dev * th.incx, word_size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (th.dim_n_dev * th.incy, word_size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(ch,th,th.dim_n_dev,alpha_r,dev_X_r,dev_Y_r);
			cublasGetVector (th.dim_n_dev, word_size_r, dev_Y_r, 1, hst_Y_r, 1);
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
			#else
			ROUTINE_REFERENCE(th,th.dim_n_dev,alpha_r,hst_X_r,hst_Y_r);
			#endif
			mublasCheckVector (th.dim_n_dev * th.incy, hst_Y_t, hst_Y_r);
		}

		if (th.mode != 't') {
			#if defined (MUBLAS)
			print_mublas_info1 (&th, &ha);
			#endif
		}
		printf ("\n");

		dim_dev_increment (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		cudaFreeHost (hst_X_r);
		cudaFreeHost (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CHECK_CUBLAS)
	cublasDestroy (ch);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &ha);
	mublasDestroy (&ha);
	#elif defined (CUBLAS)
	cublasDestroy (ha);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
// --------------------------------------------

	#endif
	return 0;
}
