#include "testing_common.h"
#include <sys/wait.h>
#define PREC_NO
#include "testing_common.cpp"
#include <gsl/gsl_integration.h>
#include <gsl/gsl_spline.h>
#define WBUFFER 10000

typedef struct {
	gsl_interp_accel *acc;
	gsl_spline *spline;
}param_t;

double func (const double x, void *param_v) {
	param_t *param = (param_t*)param_v;
	gsl_interp_accel *acc = param->acc;
	gsl_spline *spline = param->spline;
	return gsl_spline_eval (spline, x, acc);
}

int32_t
main (int32_t argc, char **argv)
{
	pid_t pid;
	uint32_t i;
	int32_t  status;
	char buf[256];
	char command[512];// = "./testing_cublas_sgemm_mgpu --numgpus=1 --range=10240:10240:1 --mode=e > test.dat_";

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
//	print_info1 (&th);
// --------------------------------------------

	i = 2;
	strcat (command, th.exec);
	strcat (command, " --nodisp=1 --mode=e");
	while (argv[i]) {
		strcat (command, " ");
		strcat (command, argv[i++]);
	}
	strcat (command, " > test.dat_");

	// mgpu
	int* devs;
	devs = (int *)malloc(sizeof(int)*th.num_installed_gpus);
	for (i=0;i<th.num_installed_gpus;i++)
		devs[i] = i;
	
	nvmlInit();
	nvmlDevice_t* nvmldevs;
	nvmldevs = (nvmlDevice_t *)malloc(sizeof(nvmlDevice_t)*th.num_installed_gpus);
	for (i=0;i<th.num_installed_gpus;i++)
		nvmlDeviceGetHandleByIndex(i, &nvmldevs[i]);
	int32_t count = 0;
	uint32_t *power;
	power = (uint32_t *)malloc(sizeof(uint32_t)*th.num_installed_gpus);
//	uint32_t *temp;
//	temp = (uint32_t *)malloc(sizeof(uint32_t)*th.num_installed_gpus);
//	nvmlDeviceGetTemperature (nvmldevs, NVML_TEMPERATURE_GPU, &temp);

	double powers[WBUFFER];
	double times[WBUFFER];
	for (i = 0; i < WBUFFER; i++) powers[i] = 0.;

	if ((pid = fork ()) < 0)
		perror ("fork error");
    else if (pid == 0) {
		system (command);
		exit (1);
	} else do {
		if ((pid = waitpid (pid, &status, WNOHANG)) == -1)
			perror ("waitpid error");
		else if (pid == 0) {
			times[count] = gettime2();
			for (i=0;i<th.num_installed_gpus;i++)
				nvmlDeviceGetPowerUsage (nvmldevs[i], &power[i]);
			for (i=0;i<th.num_installed_gpus;i++)
				powers[count] += (double)power[i]/1000;

//			printf ("%d\t%f\t%f\t", count, times[count], gettime2());
//			for (i=0;i<th.num_installed_gpus;i++) printf ("%10.4f\t", (float)power[i]/1000);
//			printf ("%10.4f\n", powers[count]);

			count++;
			usleep (15000); //15msec
		} else {
			/*
			if (WIFEXITED (status))
				printf ("%% child process exited with status of %d\n", WEXITSTATUS (status));
			else
				puts ("%% child process did not exit successfully");
			*/
		}
	}
    while (pid == 0);
	nvmlShutdown();

	FILE *fp;
	size_t mm, nn, kk;
	double x0, x1;
	double exec_time, exec_flops, exec_gbs;
	int32_t exec_iter;
	if ((fp = fopen ("test.dat_", "r")) == NULL) {
		printf ("mublas testing error: file open error.\n");
		exit (1);
	}	
	while (fgets (buf, 256, fp) != NULL) {
		sscanf (buf, "%zd\t%zd\t%zd\t%lf\t%lf\t%le\t%le\t%le\t%d", &mm, &nn, &kk, &x0, &x1, &exec_time, &exec_flops, &exec_gbs, &exec_iter);
	}
	fclose (fp);
	// exec_time from buf is the time for one execution
	exec_time = x1 - x0;

// ----- GSL integration -----
	double ret, err;
	gsl_interp_accel *acc = gsl_interp_accel_alloc();
	gsl_spline *spline = gsl_spline_alloc(gsl_interp_cspline, count);
	gsl_spline_init (spline, times, powers, count);
	gsl_integration_workspace * work = gsl_integration_workspace_alloc (WBUFFER);
	gsl_function Func;
	param_t param = {acc, spline};
	Func.function = &func;
	Func.params = &param;
	//gsl_integration_qags (&Func, x0, x1, 0, 1.0e-4, WBUFFER, work, &ret, &err);
	gsl_integration_qag (&Func, x0, x1, 0, 1.0e-4, WBUFFER, GSL_INTEG_GAUSS41, work, &ret, &err);
	gsl_integration_workspace_free(work);
	gsl_spline_free(spline);
	gsl_interp_accel_free(acc);
//	printf ("result        = %2.3e\n", ret);
//	printf ("estimated err = %2.3e\n", err);
//	printf ("intervals     = %zu\n", work->size);
// ----- GSL integration -----

/*
	printf ("exec: %f, %f\n", x0, x1);
	printf ("exec_iter: %d\n", exec_iter);
	printf ("exec_time: %1.3e\n", exec_time);
	printf ("exec_flops: %1.3e\n", exec_flops);
	printf ("exec_gbs: %1.3e\n", exec_gbs);
	printf ("W: %1.3e\n", ret);
	printf ("W/s: %1.3e\n", ret/exec_time);
	printf ("GFlops/W: %1.3e\n", exec_flops/(ret/exec_time));
*/	
	printf ("%% M\tN\tK\tTime[sec]\t#iter\t[GFlops]\t[W]\t[GFlops/W]\n");
	printf ("%zd\t%zd\t%zd\t%1.3e\t%d\t%1.3e\t%1.3e\t%1.3e\n", mm, nn, kk, exec_time, exec_iter, exec_flops, ret/exec_time, exec_flops/(ret/exec_time));
}

